/*
** GIUNIVF.g
**
** purpose : compute Generalized Impulse response function for univariate
**           time series
**
** format  : { GI }=GIUNIVF(shock_v,hist,yi,N,R,&model,&dist,&tr_res);
**
** input   : shock_v  : (kx1) vector, shocks at time t for which to compute GI
**           hist     : (1xp) vector, history of time series
**           yi       : scalar, index of dependent variable in hist
**           N        : scalar, maximum horizon for which to compute GI
**           R        : scalar, number of replications to use in resampling
**           &model   : pointer to procedure to compute new histories
**           &dist    : pointer to procedure to generate STANDARDIZED residuals
**           &tr_res  : pointer to procedure to transform residuals (adjust for variance)
**
** output  : GI       : ((N+1)xk) vector, containing GI
**
** globals :
**
** remarks : 1) The procedure &model should have two inputs:
**                (pxR) matrix of histories and 
**                (1xR) matrix of shocks, 
**           2) The single output of &model should be a (pxR) matrix of "new" histories
**           3) The procedure &dist should give an (R x k) matrix of standardized residuals
**              as the only output, where R and k are the two inputs
**
**           Any parameters used by these procedures should be defined as globals   
**
** authors : Dick van Dijk
**
** date    : First attempt on 26 February 1997
**           Last revision on 22 December 1999
**
*/
proc GIUNIVF(shock_v,hist,yi,N,R,&model,&dist,&tr_res);
  local n_sh,realvb,realzb,GI,st_shock,i,histz,shocz,histv,shocv,j;
  local model: proc;
  local dist: proc;
  local tr_res: proc;

n_sh=rows(shock_v);

/* initialization */
realvb=zeros(1,N+1);
realzb=zeros(1,N+1);
GI=zeros(n_sh,N+1);

/* generate standardized shocks */
st_shock=dist(R,N+1);

/* benchmark profile */
histz=hist.*ones(R,1);
i=0;
do until (i==N+1);
  i=i+1;
  shocz=tr_res(histz,st_shock[.,i]);
  histz=model(histz,shocz);
  realzb[1,i]=meanc(histz[.,yi]);
endo;

j=0;
do until (j==n_sh);
  j=j+1;
  shocv=shock_v[j].*ones(R,1);
  histv=hist.*ones(R,1);
  histv=model(histv,shocv);
  realvb[1,1]=meanc(histv[.,yi]);

  /* iterate over horizon */
  i=1;
  do until (i==N+1);
    i=i+1;
    shocv=tr_res(histv,st_shock[.,i]);
    histv=model(histv,shocv);
    realvb[1,i]=meanc(histv[.,yi]);
  endo;

  GI[j,.]=realvb-realzb;
endo;

retp(GI');
endp ;
