/*
** URNSAFCT.e 
** Compute forecasts of US unemployment rate from two-regime STAR model
**
** Last revision: 23 March 2000
** Written by Dick van Dijk
*/

new; cls;
__maxvec=100000;

p_max=18;                               /* Maximum autoregressive lag */
u_dif=1;                                /* u_dif=1 if model is for first differences, 
                                           u_dif=0 if model is for levels */
i_lev=1;                                /* i_lev=1 if lagged level is to be included, 
                                           i_lev=0 if not (NB only relevant if u_dif=1)*/

l_dif=12;                               /* value of "long difference" which is used as transition variable */
max_yd=6;                               /* maximum delay of transition variable */

/*
** Specify sample period of data in files
*/
n_fy=48;                                /* Year of first observation */
n_fm=1;                                 /* Month of first observation */
n_ly=99;                                /* Year of last observation */
n_lm=12;                                /* Month of last observation */
d_nobs=12*(n_ly-n_fy)+n_lm-n_fm+1;      /* Number of observations */

/*
** Specify sample period for which model has been estimated
*/
e_fy=68; /* year of first observation */
e_fm=6;  /* month of first observation */
e_ly=89; /* year of last observation */
e_lm=12; /* month of last observation */

/*
** Specify forecast period
*/
f_fy=90; /* year of first observation */
f_fm=1;  /* month of first observation */
f_ly=99; /* year of last observation */
f_lm=12; /* month of last observation */

/* Create monthly dummies */
m_dum=ones(n_ly-n_fy+1,1).*.eye(12);
m_dum=m_dum[n_fm:n_fm+d_nobs-1,.];

/* load raw data */
load lhmu[]=lhmunsa.dat; /* number of unemployed */
load lhmc[]=lhmcnsa.dat; /* labor force */
/* compute unemployment rate */
y_raw=(lhmu./lhmc)*100;

e_nobs=12*(e_ly-e_fy)+e_lm-e_fm+1; /* Number of observations in estimation period */
f_nobs=12*(f_ly-f_fy)+f_lm-f_fm+1; /* Number of observations in forecast period */

tdat=packr(y_raw[12*(e_fy-n_fy)+e_fm-n_fm+1:12*(e_fy-n_fy)+e_fm-n_fm+e_nobs+f_nobs]~
           m_dum[12*(e_fy-n_fy)+e_fm-n_fm+1:12*(e_fy-n_fy)+e_fm-n_fm+e_nobs+f_nobs,.]);
y_lev=tdat[.,1];
m_dum=tdat[.,2:13];
T=rows(y_lev);

if u_dif; /* model for first differences */
  y_dif=(y_lev[2:T]-y_lev[1:T-1]); 
  m_dum=m_dum[2:T,.];
  T=T-1;  
  if i_lev; /* include lagged level term */
    deter=ones(T,1)~m_dum~y_lev[1:T];
  else;
    deter=ones(T,1)~m_dum;
    endif;
else; /* model for levels */
  y_dif=y_lev;
  deter=ones(T,1)~m_dum~seqa(1,1,T);
endif;

/* Construct data matrix */
_dat=trimr(lagn(y_dif,seqa(0,1,p_max+1)),p_max,0);
j=0;
do until (j==max_yd);
  j=j+1;
  if u_dif;
    _dat=_dat~sumc(_dat[.,1+j:1+j+l_dif-1]');
  else;
    _dat=_dat~(_dat[.,1+j]-_dat[.,1+j+l_dif]);
  endif;
endo;
_dat=ones(T-p_max,1)~m_dum[p_max+1:T,.]~seqa(1/(T-p_max),1/(T-p_max),T-p_max)~_dat;
if i_lev;
  _dat=_dat~y_lev[p_max+1:T];
endif;
T=rows(_dat);
T_est=T-f_nobs;

/*
** Specify column index of regressand (_yi), 
** indices of regressors which enter linearly (_xi0), 
** indices of regressors which enter in "lower" regime G=0 (_xi1), 
** indices of regressors which enter in "upper" regime G=1 (_xi2), 
** and index of transition variable (_xtri),
*/
_yi=15;
_xi1=(14+1+p_max+max_yd+1)|16|21|23|25|28|30;
_xi2=(14+1+p_max+max_yd+1)|16|17|22|28|29|30;
_xi0=1|2|3|4|5|6|7|8|9|10|11;
_xtri=(14+1+p_max+1);

if not scalmiss(_xi0);
  nk_x0=rows(_xi0);
else;
  nk_x0=0;
endif;
if not scalmiss(_xi1);
  nk_x1=rows(_xi1);
else;
  nk_x1=0;
endif;
if not scalmiss(_xi2);
  nk_x2=rows(_xi2);
else;
  nk_x2=0;
endif;

_xi=_xi0|_xi1|_xi2;
nk_xi=nk_x0|nk_x1|nk_x2;
nk_x=sumc(nk_xi);
nk_xii=0|cumsumc(nk_xi);

/* column indices of constant, dummies, lags and lagged level */
_xic=indexcat(_xi,0|1.5);
_xidum=indexcat(_xi,0|12.5);
_xilag=indexcat(_xi,13.5|(14+1+p_max+max_yd+2));
_yilev=(14+1+p_max+max_yd+1);

/* compute scaling factor for transition function */
_sc=stdc(_dat[1:T_est,_xtri]);

/* load parameters */
load theta[]="LSTRPAR.out";

/* load residuals and standardize */
load _res[]="LSTRRES.out";
_sd_res=stdc(_res);
_std_res=_res./_sd_res;

/*
** h_max is the maximum forecast horizon
** R is the number of replications that are used in constructing the bootstrap forecast
*/
h_max=12;
R=2500;

/* hind_v containst row-indices of "histories" (forecast origins) for which to compute forecasts */
hind_v=seqa(T_est+1,1,f_nobs-h_max+1);
n_hist=rows(hind_v);

real_y=zeros(n_hist,h_max);
fcst_y=zeros(n_hist,h_max);

/* Compute forecasts for all histories */
tim=date;
i=0;
do until (i==n_hist);
  i=i+1;

  real_y[i,.]=(_dat[hind_v[i]:hind_v[i]+h_max-1,_yilev]+_dat[hind_v[i]:hind_v[i]+h_max-1,_yi])';
  hist=_dat[hind_v[i],.];
  
  rndseed 410124;
  {fcst_t,fcst_b}=FCSTUNIV(hist,_yilev,h_max,R,&fxlstr,&sboot,&trnorm);
  fcst_y[i,.]=fcst_t';

  info(i,n_hist,tim);

  if (i%25==0);
    dataset="URFSTAR";
    z=saved(fcst_y~real_y,dataset,0);
  endif;
endo;

dataset="URFSTAR";
z=saved(fcst_y~real_y,dataset,0);


/*
**  FXLSTR.g
**
**  procedure to compute new histories for two-regime LSTR model
**
*/
proc (1)=FXLSTR(hist,shock);
  local y,histn;

y=LSTR(hist,shock);
histn=ones(rows(hist),1)~rotater(hist[.,2:13],1)~(hist[.,14]+(1./(T-p_max)))~
   y~y~hist[.,16:p_max+14]~(sumc((y~hist[.,16:15+l_dif-1])'))~
   hist[.,14+1+p_max+1:14+1+p_max+max_yd-1]~(y+hist[.,14+1+p_max+max_yd+1]);

retp(histn);
endp;


/*
**  LSTR.g
**
**  procedure to compute fitted value for two-regime LSTR model
**
*/
proc (1)=LSTR(hist,shock);
  local fun,y;

fun=LOGF(hist[.,_xtri],_sc,theta[nk_x+1],theta[nk_x+2]);
if (nk_x0 GT 0);
  y=hist[.,_xi0]*theta[1:nk_x0] + 
     (1-fun).*hist[.,_xi1]*theta[nk_x0+1:nk_x0+nk_x1]+
      fun.*hist[.,_xi2]*theta[nk_x0+nk_x1+1:nk_x] + shock;
else;
  y=(1-fun).*hist[.,_xi1]*theta[nk_x0+1:nk_x0+nk_x1]+
        fun.*hist[.,_xi2]*theta[nk_x0+nk_x1+1:nk_x] + shock;
endif;

retp(y);
endp;


/*
**  LOGF.g
**
**  procedure to compute the value of the logistic transition function
**
*/
proc LOGF(x_trans,sc,gam,c);

retp(1./(1+exp(-(gam./sc).*(x_trans-c))));
endp;


proc(1)=STNorm(R,k);

retp(rndn(R,k));
endp;


proc(1)=TRnorm(hist,shock);

retp(shock.*_sd_res);
endp;


proc(1)=sboot(R,k);
  local ind;

ind=ceil(rndu(R*k,1)*rows(_std_res));

retp(reshape(_std_res[ind],R,k));
endp;
