/*
** DGNLSTR.g
**
** purpose : check adequacy of fitted LSTR model by computing LM-tests for
**           A) serial independence of errors
**           B) remaining nonlinearity
**           C) parameter constancy
**           see Eitrheim, O. and T. Tersvirta (1996), Testing the adequacy of
**                 smooth transition autoregressive models, Journal of Econometrics 74, 59-75.
**               van Dijk, D. and P.H. Franses (1999), Multiple regime smooth
**                 transition models, Macroeconomic Dynamics 3, 311-340.
**               Medeiros, M. and A. Veiga (2001), Diagnostic checking in a flexible nonlinear
**                 time series model, Journal of Time Series Analysis, to appear.
**
** format  : {LM_SIc,pv_SIc,LM_SIf,pv_SIf,LM_RNLc,pv_RNLc,LM_RNLf,pv_RNLf,
**            LM_Cc,pv_Cc,LM_Cf,pv_Cf,LM_NLvc,pv_NLvc,LM_NLvf,pv_NLvf,LM_Cvc,pv_Cvc,LM_Cvf,pv_Cvf} =
**            DGNLSTR(resid,theta,q_max,_xtre,_xtve,_xnle,exc_nl);
**
** input   : resid     : (T x 1) vector of residuals
**           theta     : ((nk_x+2) x 1) vector of parameters
**           q_max     : scalar, maximum order of serial correlation to test against
**           _xtre     : column indices of transition variables in additional nonlinear part
**           _xtve     : column indices of regressors to test for parameter constancy
**           _xnle     : column indices of regressors to test for remaining nonlinearity
**           exc_nl    : exclusion matrix, determining which elements of gradient to exclude in
**                          auxiliary regressions (to avoid singularity of moment matrices)
**
** output  : LM_SI.    : (q_max x 1), statistics for testing no serial dependence
**           pv_SI.    : (q_max x 1), corresponding p-values
**           LM_RNL.   : (k_xtre x 1), statistics for testing no remaining nonlinearity
**           pv_RNL.   : (k_trax1), corresponding p-values
**           LM_C.     : scalars, statistics for testing parameter-constancy
**           pval_C.   : scalars, corresponding p-values
**           LM_NLv.   : (k_xtre x 1), statistics for testing no nonlinearity in variance
**           pv_NLv.   : (k_trax1), corresponding p-values
**           LM_Cv.    : scalars, statistics for testing constancy of variance
**           pval_Cv.  : scalars, corresponding p-values
**
** globals : none
**
** remarks : Both chi-square- (C) and F-variants (f) of statistics are computed
**
** authors : Dick van Dijk
**
** date    : First attempt on 25 March 1996
**           Last revision on  5 May   1999
*/
proc (20)=DGNLSTR(resid,theta,q_max,_xtre,_xtve,_xnle,exc_nl);
  local fun,gradc,n_obs,orth_par,res_0,SSR_0,q_v,LM_SIc,
        pv_SIc,LM_SIf,pv_SIf,i,q,lagres,aux_m,res,LM_Cc,pv_Cc,LM_Cf,pv_Cf,
		nk_xtv,maxt,t_mat,k_xtre,LM_RNLc,pv_RNLc,LM_RNLf,pv_RNLf,
		nk_e,tgradc,nk_xnl,LM_NLvc,pv_NLvc,LM_NLvf,pv_NLvf,LM_Cvc,pv_Cvc,LM_Cvf,pv_Cvf;

if _thrzero;
  theta=theta[1]|0;
endif;
fun=LOGF(_dat[.,_xtri],_sc,exp(theta[1]),theta[2]);
gradc=GETXXX(_dat,fun);
n_obs=rows(gradc);
orth_par=invpd(moment(gradc,0))*(gradc'resid);
res_0=resid-gradc*orth_par;
SSR_0=res_0'res_0;

/* compute LM_SIs and corresponding pvalues */
q_v=seqa(1,1,q_max);
LM_SIc=zeros(rows(q_v),2);
pv_SIc=zeros(rows(q_v),2);
LM_SIf=zeros(rows(q_v),2);
pv_SIf=zeros(rows(q_v),2);
i=0;
do until (i==rows(q_v));
  i=i+1;
  q=q_v[i];
  lagres = seqa(1,1,q);
  aux_m = trimr(gradc~rgr(res_0,lagres),q,0);
  res = trimr(res_0,q,0);
  {LM_SIc[i,1],pv_SIc[i,1],LM_SIf[i,1],pv_SIf[i,1]}=CMPLMT(res,nk_x,aux_m,q);
  {LM_SIc[i,2],pv_SIc[i,2],LM_SIf[i,2],pv_SIf[i,2]}=
      CMPLMTW(res,aux_m[.,1:nk_x],aux_m[.,nk_x+1:nk_x+q]);
endo;

LM_Cc=zeros(3,2);
pv_Cc=zeros(3,2);
LM_Cf=zeros(3,2);
pv_Cf=zeros(3,2);

/* create matrix containing t, t^2 and t^3, t=1,...,T */
maxt = 3 ;
nk_xtv=rows(_xtve);
t_mat = seqa(1,1,n_obs).^(seqa(1,1,maxt))' ;

/* create "full" regressor matrix for LM_C1 */
aux_m = gradc~(gradc[.,_xtve].*t_mat[.,1]);

/* compute LM_C1 and corresponding pvalue */
{LM_Cc[1,1],pv_Cc[1,1],LM_Cf[1,1],pv_Cf[1,1]}=CMPLMT(res_0,nk_x,aux_m,nk_xtv);
{LM_Cc[1,2],pv_Cc[1,2],LM_Cf[1,2],pv_Cf[1,2]}=
    CMPLMTW(res_0,aux_m[.,1:nk_x],aux_m[.,nk_x+1:nk_x+nk_xtv]);

/* create "full" regressor matrix for LM_C2 */
aux_m = aux_m~(gradc[.,_xtve].*t_mat[.,2]);

/* compute LM_C2 and corresponding pvalue */
{LM_Cc[2,1],pv_Cc[2,1],LM_Cf[2,1],pv_Cf[2,1]}=
    CMPLMT(res_0,nk_x,aux_m,2*nk_xtv);
{LM_Cc[2,2],pv_Cc[2,2],LM_Cf[2,2],pv_Cf[2,2]}=
    CMPLMTW(res_0,aux_m[.,1:nk_x],aux_m[.,nk_x+1:nk_x+2*nk_xtv]);

/* create "full" regressor matrix for LM_C3 */
aux_m = aux_m~(gradc[.,_xtve].*t_mat[.,3]);

/* compute LM_C3 and corresponding pvalue */
{LM_Cc[3,1],pv_Cc[3,1],LM_Cf[3,1],pv_Cf[3,1]}=
    CMPLMT(res_0,nk_x,aux_m,3*nk_xtv);
{LM_Cc[3,2],pv_Cc[3,2],LM_Cf[3,2],pv_Cf[3,2]}=
    CMPLMTW(res_0,aux_m[.,1:nk_x],aux_m[.,nk_x+1:nk_x+3*nk_xtv]);

LM_Cvc=zeros(3,2);
pv_Cvc=zeros(3,2);
LM_Cvf=zeros(3,2);
pv_Cvf=zeros(3,2);
{LM_Cvc[1,1],pv_Cvc[1,1],LM_Cvf[1,1],pv_Cvf[1,1]}=CMPLMT((res_0.^2)./(SSR_0/T)-1,1,t_mat[.,1],1);
{LM_Cvc[1,2],pv_Cvc[1,2],LM_Cvf[1,2],pv_Cvf[1,2]}=
    CMPLMTW((res_0.^2)./(SSR_0/n_obs)-1,ones(n_obs,1),t_mat[.,1]);

{LM_Cvc[2,1],pv_Cvc[2,1],LM_Cvf[2,1],pv_Cvf[2,1]}=CMPLMT((res_0.^2)./(SSR_0/n_obs)-1,1,t_mat[.,1:2],2);
{LM_Cvc[2,2],pv_Cvc[2,2],LM_Cvf[2,2],pv_Cvf[2,2]}=
    CMPLMTW((res_0.^2)./(SSR_0/n_obs)-1,ones(n_obs,1),t_mat[.,1:2]);

{LM_Cvc[3,1],pv_Cvc[3,1],LM_Cvf[3,1],pv_Cvf[3,1]}=CMPLMT((res_0.^2)./(SSR_0/n_obs)-1,1,t_mat[.,1:3],3);
{LM_Cvc[3,2],pv_Cvc[3,2],LM_Cvf[3,2],pv_Cvf[3,2]}=
    CMPLMTW((res_0.^2)./(SSR_0/n_obs)-1,ones(n_obs,1),t_mat[.,1:3]);

k_xtre=rows(_xtre);
LM_RNLc=zeros(3*k_xtre,2);
pv_RNLc=zeros(3*k_xtre,2);
LM_RNLf=zeros(3*k_xtre,2);
pv_RNLf=zeros(3*k_xtre,2);

LM_NLvc=zeros(3*k_xtre,2);
pv_NLvc=zeros(3*k_xtre,2);
LM_NLvf=zeros(3*k_xtre,2);
pv_NLvf=zeros(3*k_xtre,2);

nk_xnl=rows(_xnle);
nk_e=sumc(exc_nl');
i=0;
do until (i==k_xtre);
  i=i+1;
  /* create matrix containing _dat[.,xtre[.,i]], _dat[.,xtre[.,i]]^2
     and _dat[.,xtre[.,i]]^3 */
  maxt = 3 ;
  t_mat = _dat[.,_xtre[i]].^(seqa(1,1,maxt))' ;

  /* create "full" regressor matrix for LM_RNL1 */
  tgradc=delif(gradc[.,_xnle]',exc_nl[i,.]')';
  aux_m = gradc~(tgradc.*t_mat[.,1]);

  /* compute LM_RNL1 and corresponding pvalue */
  {LM_RNLc[i,1],pv_RNLc[i,1],LM_RNLf[i,1],pv_RNLf[i,1]}=
      CMPLMT(res_0,nk_x,aux_m,nk_xnl-nk_e[i]);
  {LM_RNLc[i,2],pv_RNLc[i,2],LM_RNLf[i,2],pv_RNLf[i,2]}=
      CMPLMTW(res_0,aux_m[.,1:nk_x],aux_m[.,nk_x+1:nk_x+nk_xnl-nk_e[i]]);

  /* create "full" regressor matrix for LM_RNL2 */
  aux_m = aux_m~(tgradc.*t_mat[.,2]);

  /* compute LM_RNL2 and corresponding pvalue */
  {LM_RNLc[k_xtre+i,1],pv_RNLc[k_xtre+i,1],LM_RNLf[k_xtre+i,1],pv_RNLf[k_xtre+i,1]}=
      CMPLMT(res_0,nk_x,aux_m,2*(nk_xnl-nk_e[i]));
  {LM_RNLc[k_xtre+i,2],pv_RNLc[k_xtre+i,2],LM_RNLf[k_xtre+i,2],pv_RNLf[k_xtre+i,2]}=
      CMPLMTW(res_0,aux_m[.,1:nk_x],aux_m[.,nk_x+1:nk_x+2*(nk_xnl-nk_e[i])]);

  /* create "full" regressor matrix for LM_RNL3 */
  aux_m = aux_m~(tgradc.*t_mat[.,3]);

  /* compute LM_RNL3 and corresponding pvalue */
  {LM_RNLc[2*k_xtre+i,1],pv_RNLc[2*k_xtre+i,1],LM_RNLf[2*k_xtre+i,1],pv_RNLf[2*k_xtre+i,1]}=
      CMPLMT(res_0,nk_x,aux_m,3*(nk_xnl-nk_e[i]));
  {LM_RNLc[2*k_xtre+i,2],pv_RNLc[2*k_xtre+i,2],LM_RNLf[2*k_xtre+i,2],pv_RNLf[2*k_xtre+i,2]}=
      CMPLMTW(res_0,aux_m[.,1:nk_x],aux_m[.,nk_x+1:nk_x+3*(nk_xnl-nk_e[i])]);

  {LM_NLvc[i,1],pv_NLvc[i,1],LM_NLvf[i,1],pv_NLvf[i,1]}=
      CMPLMT((res_0.^2)./(SSR_0/n_obs)-1,1,t_mat[.,1],1);
  {LM_Nlvc[i,2],pv_NLvc[i,2],LM_NLvf[i,2],pv_NLvf[i,2]}=
      CMPLMTW((res_0.^2)./(SSR_0/n_obs)-1,ones(n_obs,1),t_mat[.,1]);

  {LM_NLvc[k_xtre+i,1],pv_NLvc[k_xtre+i,1],LM_NLvf[k_xtre+i,1],pv_NLvf[k_xtre+i,1]}=
      CMPLMT((res_0.^2)./(SSR_0/n_obs)-1,1,t_mat[.,1:2],2);
  {LM_NLvc[k_xtre+i,2],pv_NLvc[k_xtre+i,2],LM_NLvf[k_xtre+i,2],pv_NLvf[k_xtre+i,2]}=
      CMPLMTW((res_0.^2)./(SSR_0/n_obs)-1,ones(n_obs,1),t_mat[.,1:2]);

  {LM_NLvc[2*k_xtre+i,1],pv_NLvc[2*k_xtre+i,1],LM_NLvf[2*k_xtre+i,1],pv_NLvf[2*k_xtre+i,1]}=
      CMPLMT((res_0.^2)./(SSR_0/n_obs)-1,1,t_mat[.,1:3],3);
  {LM_NLvc[2*k_xtre+i,2],pv_NLvc[2*k_xtre+i,2],LM_NLvf[2*k_xtre+i,2],pv_NLvf[2*k_xtre+i,2]}=
      CMPLMTW((res_0.^2)./(SSR_0/n_obs)-1,ones(n_obs,1),t_mat[.,1:3]);

endo;

retp(LM_SIc,pv_SIc,LM_SIf,pv_SIf,LM_RNLc,pv_RNLc,LM_RNLf,pv_RNLf,
     LM_Cc,pv_Cc,LM_Cf,pv_Cf,LM_NLvc,pv_NLvc,LM_NLvf,pv_NLvf,LM_Cvc,pv_Cvc,LM_Cvf,pv_Cvf);
endp;

proc(4)=CMPLMT(res_0,k_0,aux_m,k_restr);
  local stdv,stdvzero,cind,T,SSR_0,invaux_m,phi_1,res_1,SSR_1,LM_stc,
        LM_pvc,LM_stf,LM_pvf;

T=rows(res_0);
res_1=1;
SSR_0=res_0'res_0;

invaux_m=invpd(moment(aux_m,0));
if not scalmiss(invaux_m);
  phi_1 = invaux_m*(aux_m'res_0);
  res_1 = res_0 - aux_m*phi_1;
  SSR_1 = res_1'res_1;

  LM_stc = T*(SSR_0 - SSR_1)/SSR_0;
  LM_pvc = CDFCHIC(LM_stc,k_restr);
  LM_stf = ((SSR_0 - SSR_1)/k_restr)/(SSR_1/(T-k_restr-k_0));
  LM_pvf = CDFFC(LM_stf,k_restr,T-k_restr-k_0);
else;
  LM_stc=0;
  LM_pvc=0;
  LM_stf=0;
  LM_pvf=0;
endif;

retp(LM_stc,LM_pvc,LM_stf,LM_pvf);
endp;


proc(4)=CMPLMTW(y,x_0,x_restr);
  local stdv,stdvzero,cind,T,k_0,k_restr,c,inv_m,phi_0,res_0,phi_1,res_1,
        rr,LM_stc,LM_pvc,LM_stf,LM_pvf;

T=rows(y);
k_0=cols(x_0);
k_restr=cols(x_restr);

c=ones(T,1);

inv_m=invpd(moment(x_0,0));
if not scalmiss(inv_m);
  phi_0 = inv_m*(x_0'y);
  res_0 = y - x_0*phi_0;
  phi_1 = inv_m*(x_0'x_restr);
  res_1 = x_restr - x_0*phi_1;
  rr=res_0.*res_1;
  LM_stc=(c'rr*invpd(rr'rr)*rr'c);
  LM_pvc=CDFCHIC(LM_stc,k_restr);
  LM_stf=(LM_stc/k_restr)/((T-LM_stc)/(T-k_restr-k_0));
  LM_pvf = CDFFC(LM_stf,k_restr,T-k_restr-k_0);
else;
  LM_stc=0;
  LM_pvc=0;
  LM_stf=0;
  LM_pvf=0;
endif;

retp(LM_stc,LM_pvc,LM_stf,LM_pvf);
endp;
