/*
** UKERNEL.g
**
** purpose : calculate a univariate kernel (density) estimate
**           and its derivative
**
** format  : { f,d,h }=UKERNEL(xs,x,h,w,&kf);
**
** input   : xs   : (nsx1) vector, points where density is to be estimated
**           x    : (nx1) vector, observed data points
**           h    : (kx1) vector, bandwidths
**           w    : (nx1) vector, weights for different observations
**           &kf  : pointer to procedure containing kernel function
**
** output  : f    : (nsxk) matrix, estimated density in xs, each column in f
**                  corresponds to a bandwidth
**           d    : (nsxk) matrix, derivative of estimated density
**           h    : (kx1) vector, bandwidths
**
** globals :
**
** remarks : If h<=0, the bandwidth is determined by the procedure bandws(x)
**
** authors : Dick van Dijk
**
** date    : First attempt on  7 October  1996
**           Last revision on 17 December 1997
**
*/
proc (3)=ukernel(xs,x,h,w,&kf);
  local ns,k,f,d,sw,i,arg,kff,kfd;
  local kf:proc;

/* Some error checking here */
if (cols(xs)>1);
  errorlog "UKERNEL.g: xs has too many columns";
  retp(-1,-1,-1 );
endif;
if (cols(x)>1);
  errorlog "UKERNEL.g: x has too many columns";
  retp( -1,-1,-1 );
endif;

/* Determine bandwidth if necessary */
if (h<=0);
  h=bandws(x);
endif;

/* Initialization */
ns=rows(xs);
k=rows(h);
f=zeros(ns,k);
d=zeros(ns,k);

sw=sumc(w);
/* Compute density estimates */
i=0;
do until (i==ns);
  i=i+1;
  arg=(x-xs[i])./h';
  {kff,kfd}=kf(arg);
  f[i,.]=(sumc(kff.*w)./(sw.*h))';
  d[i,.]=(sumc(kfd.*w)./(sw.*(h.^2)))';
endo;

retp(f,d,h);
endp;


/*
** BANDWS.g
**
** purpose : calculate the optimal bandwidth in kernel estimation of a density
**           by using the rule of thumb from Silverman (1986)
**
** format  : { h }=BANDWS(x);
**
** input   : x   : (nx1) vector, series for which density is to be estimated
**
** output  : h   : scalar, bandwidth
**
** globals :
**
** remarks : The optimal bandwidth is calculated according to
**           eq. 3.31 of Silverman (1986)
**
** authors : Dick van Dijk
**
** date    : First attempt on  7 October  1996
**           Last revision on 15 February 1997
**
*/
proc bandws(x);
  local n,xs,qi1,qi3,iqr,s;

/* Some error-checking here */
if (cols(x)>1);
   errorlog "input error in BANDWS.g: too many columns";
retp(-1);
endif;

n=rows(x);
xs=sortc(x,1);
qi1=round(0.25*n);
qi3=round(0.75*n);
iqr=xs[qi3]-xs[qi1];
s=stdc(xs);

retp( 0.9*minc(s|(iqr/1.349))/n^0.2 );
endp;

proc (2)=k_gauss(u);
  retp(pdfn(u),-u.*pdfn(u));
endp;


/*
** VIEWUKNL.g
**
** purpose : plot kernel density estimate
**
** format  : call viewuknl(x,f,h,m,s);
**
** input   : x   : (nx1) vector with data points
**           f   : (nx1) vector with estimated density
**           h   : scalar, bandwidth
**           m   : scalar, mean of normal distribution to be plotted
**           s   : scalar, standard deviation of normal distribution
** output  : plot of y (horizontally) by its est'd density f(y) (vertically).
**           Whiskers can also be plotted along horizontal axis for each datum
**
** globals : All globals of the pgraph library
**           _whiskr= -1 : draw whiskers at 1/15th size;
**                    >0 : draw whiskers at size _whiskr;
**                     0 : do not draw whiskers
**           Whiskers are always drawn at y==0, so you may have to
**           use ytics to scale the y-axis.
**           _jitter=  0 : nothing extra.
**                     # : add # jitter to each whisker if _whiskr/=0
**           _normal=  0 : nothing extra
**                    >0 : draw normal distribution with mean m and
**                          standard deviation s
**           _legend=  0 : no legend
**                    >0 : legend with bandwidth
**
** remarks : Library pgraph must be activated
**
** authors : Dick van Dijk
**
** date    : First attempt on  7 October 1996
**           Last revision on 15 February  1997
**
*/
declare matrix _whiskr ?=-1;
declare matrix _jitter ?=0;
declare matrix _normal ?=0;
declare matrix _legend ?=0;

proc (0)=viewuknl(x,f,h,m,s);
  local data,k,xlow,xhigh,flow,fhigh,xlegend,ylegend,std,y,os,norm,temp;

/* Some error checking here */
if (cols(f) ne 1) or (rows(h) ne 1);
  errorlog "VIEWUKNL.g: cols(f) or rows(h) not equal to 1";
  retp();
endif;

/* set global variables pgraph */
_pdate="";
_pcolor=15;

if _normal>0;
  norm=pdfn((x-m)./s)./s;
  temp=f;
  f=maxc(f'|norm');
endif;

scale(x,f);
xlow=minc(x);
xhigh=maxc(x);
flow=minc(f);
fhigh=maxc(f);
xlegend=xlow+.8*(xhigh-xlow);
ylegend=flow+.8*(fhigh-flow);
if _legend;
  _plegctl=1|4|xlegend|ylegend;
  _plegstr="h=" $+ ftos(h,"%*.*lf",5,3) $+ "\0Normal";
endif;

if _normal>0;
  f=temp;
endif;
data=sortc(x~f,1);
y=data[.,1];
std=rndu(rows(y),1)*_jitter-(_jitter/2);
y=y+std;

if _whiskr==-1;
  if (fhigh GT 8);
    std=minc(5|floor(fhigh/5));
    flow=-ceil(fhigh/(15*std))*std;
    fhigh=ceil(fhigh);
    ytics(flow,fhigh,std,std-1);
  else;
    flow=-.1*(2^ceil(log(fhigh)/log(2)));
    fhigh=ceil(10*fhigh)/10;
    std=-flow/(2-(flow GE -.1));
    ytics(flow,fhigh,std,1);
  endif;
  os=ones(rows(y),1);
  _pline=os~(os*6)~y~os*flow~y~os*(flow+maxc(f)/15)~os~(os*15)~(os*0);
  clear os;
elseif _whiskr>0;
  if (fhigh GT 8);
    std=minc(5|floor(fhigh/5));
    flow=-ceil(fhigh/(15*std))*std; 
    fhigh=ceil(fhigh);
    ytics(flow,fhigh,std,std-1);
  else;
    flow=
      -maxc(maxc(ceil((log(fhigh)/log(2)+1))|.5)/10|ceil(10*_whiskr)/10);
    fhigh=ceil(10*fhigh)/10;
    std=maxc(-flow/2|.05);
    ytics(flow,fhigh,std,1);
  endif;
  os=ones(rows(y),1);
  _pline=os~(os*6)~y~(os*flow)~y~(os*(flow+_whiskr))~os~(os*15)~(os*0);
  clear os;
endif;

if _normal>0;
  norm=pdfn((data[.,1]-m)./s)./s;
  xy(data[.,1],data[.,2:cols(data)]~norm);
else;
  xy(data[.,1],data[.,2:cols(data)]);
endif;

endp;
