/*
** ROBUST.prc
** Procedures for robust estimation
**
** Procedure  Purpose                                                    Line
** ==========================================================================
** GM_EST     HBP-GM estimation of linear regression models                29
** M_EST      M estimation of linear regression models                    130
** LMS        Projection variant of least median of squares               203
** MVE        Compute the MVE distances (or Mahalanobis distances)		  271
**            using projection algorithm suggested in Rousseeuw and
**            van Zomeren (1990)
** MADC       Compute median absolute deviation                           354
**
** Psi-functions:
** psi_ols    OLS psi function                                            361
** psi_sgn    sign psi function (for LAD estimation)					  367
** psi_hub    Huber psi function										  373
** psi_tuk    Tukey bisquare psi function								  382
** psi_tan    bounded non-redescending psi function						  391
** psi_stu    weakly redescending psi function: Student t(5)			  397
** psi_pol	  strongly redescending psi function						  403
** psi_hubx   Huber psi function for regressors-weights					  433
** psi_polx	  strongly redescending psi function for regressor-weights	  442
**
*/


/*
** GM_EST.g
**
** purpose : HBP-GM estimation of linear regression model
**
** format  : {b,wx,we,ifail}=GM_EST(y,x,b0,&psi_x,&psi_e,maxit,tol,deter);
**
** input   : y     : (Tx1) vector, observations on regressand
**           x     : (Txk) matrix, observations on regressors
**           b0    : (kx1) vector, starting values for parameter estimates
**           psi_x : procedure to compute x-weights, should be either
**           psi_e : procedure to compute e-weights, should be either
**                    - psi_ols: OLS-weights
**                    - psi_tan:
**                    - psi_stu: Student t-weights
**                    - psi_pol:
**           maxit : scalar, maximum number of iterations
**           tol   : scalar, tolerance to be used to check for convergence
**                           of the GM-estimator
**           deter : (k_dx1) indicating indices of x which contain
**                   deterministic regressors, like constant and dummies
**
** output  : b     : GM-estimates of parameters
**           wx    : (Tx1) vector, weights for regressors
**           we    : (Tx1) vector, weights for residuals obtained from
**                                 final estimates
**           ifail : 0/1-variable, indicating whether or not convergence has
**                   been achieved within specified max number of iterations
**
** globals :
**
** remarks :
**
** authors : Andre Lucas and Dick van Dijk
**
** date    : First attempt on  3 January  1996
**           Last revision on 10 September 1997
**
*/
proc (4)=GM_EST(y,x,b0,&psi_x,&psi_e,maxit,tol,deter);
  local alpha,m,v,p,xi,di,xdi,mdist,mdist0,u_one,b,wx,we,e,s,mine,maxe,dff,iter;
  local psi_x: proc;
  local psi_e: proc;

alpha=2;
iter=0;
dff=1;
mine=-750;
maxe=750;

p=cols(x);
xi=seqa(1,1,p);
di=sumc(xi'.==deter);
xdi=delif(xi,di);
if (rows(xdi)==1);
  m=median(x[.,xdi]);
  v=madc(x[.,xdi]);
  v=v*v;
  mdist=sqrt((x[.,xdi]-m).*((x[.,xdi]-m)./v));
else;
  mdist=mve(x[.,xdi]);
endif;

wx=(psi_x(mdist.^alpha,alpha,rows(xdi))+(abs(mdist).<1e-5))./
        ((abs(mdist).<1e-5)+mdist.^alpha);
iter=0;
dff=1;
do while (iter<=maxit) and (dff>tol);
  e=y-x*b0;
  s=madc(e);
  e=e./(s.*(wx+(1e-5.*(wx.<1e-5))));
  e=substute(e,e.>maxe,maxe);
  e=substute(e,e.<mine,mine);
  we=(psi_e(e)+(abs(e).<1e-5))./(e+(abs(e).<1e-5));
  v=x'(we.*x);
  b=invpd(v)*x'(we.*y);
  dff=b-b0;
  dff=sqrt(dff'v*dff)/s;
  b0=b;
  iter=iter+1;
endo;

e=y-x*b0;
s=madc(e);
e=e./(s.*(wx+(1e-5.*(wx.<1e-5))));
e=substute(e,e.>maxe,maxe);
e=substute(e,e.<mine,mine);
if not ismiss(e);
  we=(psi_e(e)+(abs(e).<1e-5))./(e+(abs(e).<1e-5));
  if (iter>maxit);
    retp(b,wx,we,1);
  else;
    retp(b,wx,we,0);
  endif;
else;
  retp(-1,-1,-1,-1);
endif;

endp;


/*
** M_EST.g
**
** purpose : M-estimation of linear regression models
**
** format  : {b,w,ifail}=M_EST(y,x,b0,&psi,maxit,tol);
**
** input   : y     : (Tx1) vector, observations on regressand
**           x     : (Txk) matrix, observations on regressors
**           b0    : (kx1) vector, starting values
**           psi   : procedure to compute psi-weights, should be either
**                    - psi_ols: OLS-weights
**                    - psi_tan:
**                    - psi_stu: Student t-weights
**                    - psi_pol:
**           maxit : scalar, maximum number of iterations
**           tol   : scalar, tolerance to be used to check for convergence
**                           of the M-estimator
**
** output  : b     : M-estimates of parameters
**           w     : (Tx1) vector, weights for observations obtained from
**                                 final estimates
**           ifail : 0/1-variable, indicating whether or not convergence has
**                   been achieved within specified max number of iterations
**
** globals :
**
** remarks :
**
** authors : Andre Lucas and Dick van Dijk
**
** date    : First attempt on  3 January 1996
**           Last revision on 31 January 1996
**
*/
proc (3)=M_EST(y,x,b0,&psi,maxit,tol);
  local b,v,w,e,s,dff,iter,mine,maxe;
  local psi: proc;

iter=0;
dff=1;
mine=-750;
maxe=750;

do while (iter<=maxit) and (dff>tol);
  e=y-x*b0;
  s=madc(e);
  e=e./s;
  e=substute(e,e.>maxe,maxe);
  e=substute(e,e.<mine,mine);
  w=(psi(e)+(abs(e).<1e-5))./(e+(abs(e).<1e-5));
  v=x'(w.*x);
  b=invpd(v)*x'(w.*y);
  dff=b-b0;
  dff=sqrt(dff'v*dff)/s;
  b0=b;
  iter=iter+1;
endo;

e=y-x*b0;
s=madc(e);
e=e./s;
w=(psi(e)+(abs(e).<1e-5))./(e+(abs(e).<1e-5));

if (iter>maxit);
  retp(b,w,1);
else;
  retp(b,w,0);
endif;

endp;


/*
** LMS.g
**
** purpose : projection variant of least median of squares
**
** format  :  { b }=LMS(y,x);
**
** input   : y   : (Tx1) vector, observations on regressand
**           x   : (Txk) matrix, observations on regressors
**
** output  : b   : LMS-estimate of parameters
**
** globals :
**
** remarks : If the matrix x contains a constant, it should be in the first
**           column
**
** authors : Andre Lucas and Dick van Dijk
**
** date    : First attempt on  3 January 1996
**           Last revision on 18 January 1996
**
*/
proc (1)=LMS(y,x);
  local b,f0,mds,i1,dst,m1,s1,dst0;

b=y/x;
f0=median((y-x*b).*(y-x*b));
x=y~x;
mds=median(x);

i1=1;
do while i1<=rows(y);
  dst=x*(x[i1,.]'-mds);
  m1=median(dst);
  s1=1.4826*median(abs(dst-m1));
  if (s1 /=0);
    dst=(dst-m1)/s1;

    if (i1==1);
      dst0=dst;
    else;
      dst0=maxc((dst0~dst)');
    endif;
  elseif (i1==1);
    dst0=zeros(rows(dst),1);
  endif;
  i1=i1+1;
endo;

dst=(dst0 .<=sqrt(cdfchii(0.95,cols(x))));
x=x[.,2:cols(x)];

if (dst /=0);
  dst=invpd(x'(dst.*x))*x'(y.*dst);
else;
  dst=b;
endif;

if (median((y-x*dst).*(y-x*dst)) < f0);
  retp(dst);
else;
  retp(b);
endif;

endp;


/*
**  MVE.g
**
**  purpose : Compute MVE distances (or Mahalanobis distances) using the
**            projection algorithm suggested in Rousseeuw and van Zomeren,
**            JASA 1990.
**
**  format  : {dist}=MVE(xdata);
**
**  input   : xdata: NxP matrix containing the data. N cases of P dimensional
**                   variables
**
**  output  : dist: the computed Mahalanobis distances
**
**  globals :
**
**  remarks :
**
**  authors : Andre Lucas
**
**  date    : First attempt on 23 November 1992
**            Last revision on 5 January 1994
**
*/
proc (1)=mve(xdata);
  local xxdata,n,p,h,scale,loc,aid0,aid1,aid2,i,j,u,inds,meds;

/* initialize */
N=rows(xdata);
P=cols(xdata);
if (N<P);
  errorlog " MVE.g: There are less observations than regressors";
  retp(0);
endif;
meds=median(xdata)';
h=floor((n+p+1)/2);
h=floor(n/2);

if (h==0);
  errorlog "MVE.g : too few observations";
  retp(0);
endif;

/* project on x-med */
inds=seqa(1,1,n-h);
u=zeros(n,1);
i=0;
do until i==n;
  i=i+1;
  /** project the data and compute length of shortest half **/
  aid0=xdata*(xdata[i,.]-meds)';
  aid1=sortc(aid0,1);
  aid2=aid1[inds+h] - aid1[inds];
  j=minindc(aid2);
  scale=aid2[j];
  scale=maxc(scale|1e-5);
  /** compute the midpoint **/
  loc=(aid1[inds[j]+h] + aid1[inds[j]])/2;
  u=maxc((u')|(abs((aid0-loc)/scale)'));
endo;

/* project on e(i) */
xxdata=eye(p);
i=0;
do until (i==p);
  i=i+1;
  /** project the data and compute length of shortest half **/
  aid0=xdata*xxdata[.,i];
  aid1=sortc(aid0,1);
  aid2=aid1[inds+h] - aid1[inds];
  j=minindc(aid2);
  scale=aid2[j];
  scale=maxc(scale|1e-5);

  /** compute the midpoint **/
  loc=(aid1[inds[j]+h] + aid1[inds[j]])/2;
  u=maxc((u')|(abs((aid0-loc)/scale)'));
endo;

retp(u);
endp;


proc (1)=madc(x);
  /* median absolute deviation, rescaled for consistency at normal */

retp(1.483*median(abs(x-median(x)')));
endp;


proc psi_ols(x); /* ols psi functie */

retp(x);
endp;


proc (1)=psi_sgn(x); /* sign psi function, to obtain LAD-estimator*/

retp((x .GT 0) - (x .LT 0));
endp;


proc (1)=psi_hub(x); /* Huber psi function with tuning constant c=1.345*/
  local c;

c=seqa(1.345,0,rows(x));

retp(median(-c'|x'|c'));
endp;


proc (1)=psi_tuk(x); /* Tukey bisquare psi function with tuning constant c=4.685 */
  local c;

c=4.685;

retp((x.*(1-(x/c).^2).^2).*(abs(x) .LE c));
endp;


proc psi_tan(x); /* bounded non-redescending psi function */

retp(1.25*tanh(x/1.25));
endp;


proc psi_stu(x); /* weakly redescending psi function: Student t(5) */

retp(x./(1+x.*x/5));
endp;


proc psi_pol(x);
/* strongly redescending psi function; smooth version: full acceptance between
   0.01 and 0.99 quantiles, full rejection outside 0.001 and 0.999 quantiles,
   partial weighting inbetween. */
local k1,k2,a,b,c,y;

k2=1;
k1=cdfchii(1-k2/100,1);
k2=cdfchii(1-k2/1000,1);
k1=sqrt(k1);
k2=sqrt(k2);

a={}; b={};
a=a|(k1.^seqa(0,1,6)'); b=b|(k1);
a=a|((a[rows(a),.]/k1).*seqa(0,1,6)'); b=b|(1);
a=a|((a[rows(a),.]/k1).*seqa(-1,1,6)'); b=b|(0);
a=a|(k2.^seqa(0,1,6)'); b=b|(0);
a=a|((a[rows(a),.]/k2).*seqa(0,1,6)'); b=b|(0);
a=a|((a[rows(a),.]/k2).*seqa(-1,1,6)'); b=b|(0);

c=inv(a)*b;

y=abs(x);
y=(y.<=k1).*(y) + (y.<=k2).*(y.>k1).*(y.^seqa(0,1,6)'*c);
y=y.*(2*(x.>0)-1);

retp(y);
endp;


proc (1)=psi_hubx(x,alpha,p); /* Huber psi function with tuning constant c */
    local c;

c=cdfchii(0.95,p)^(alpha/2);

retp(median(-seqa(c,0,rows(x))'|x'|seqa(c,0,rows(x))'));
endp;


proc psi_polx(x,alpha,p);
/* strongly redescending psi function; smooth version: full acceptance between
   0.01 and 0.99 quantiles, full rejection outside 0.001 and 0.999 quantiles,
   partial weighting inbetween. */
  local k1,k2,a,b,c;

k2=1;
k1=cdfchii(1-k2/100,p);
k2=cdfchii(1-k2/1000,p);
k1=(k1)^(alpha/2);
k2=(k2)^(alpha/2);

a={}; b={};
a=a|(k1.^seqa(0,1,6)'); b=b|(k1);
a=a|((a[rows(a),.]/k1).*seqa(0,1,6)'); b=b|(1);
a=a|((a[rows(a),.]/k1).*seqa(-1,1,6)'); b=b|(0);
a=a|(k2.^seqa(0,1,6)'); b=b|(0);
a=a|((a[rows(a),.]/k2).*seqa(0,1,6)'); b=b|(0);
a=a|((a[rows(a),.]/k2).*seqa(-1,1,6)'); b=b|(0);

c=inv(a)*b;

x=(x.<=k1).*(x) + (x.<=k2).*(x.>k1).*(x.^seqa(0,1,6)'*c);

retp(x);
endp;

